//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

use crate::s5;

pub struct Window {
    size: s5::Rect,
    entities: Vec<Entity>,
}

impl Window {
    pub fn new(size: s5::Rect) -> Self {
        Self {
            size,
            entities: Vec::new(),
        }
    }

    pub fn size(&self) -> &s5::Rect {
        &self.size
    }

    pub fn resize(
        &mut self,
        size: s5::Rect,
    ) {
        self.size = size;
    }

    pub fn add_entity(
        &mut self,
        mut entity: Entity,
    ) {
        entity.stay_inside(&self.size);
        self.entities.push(entity);
    }

    pub fn redraw(
        &self,
        screen: &mut s5::Screen,
    ) {
        screen.clear(s5::Color {
            r: 100,
            g: 100,
            b: 100,
        });
        for entity in self.entities.iter() {
            entity.draw(screen);
        }
    }

    pub fn animate(
        &mut self,
        dt: f64,
    ) -> bool {
        for entity in self.entities.iter_mut() {
            entity.animate(&self.size, dt);
            entity.stay_inside(&self.size);
        }
        true // go on
    }

    pub fn click(
        &mut self,
        mouse: s5::Point,
    ) {
        println!("click at {:?}", mouse);
        for entity in self.entities.iter_mut() {
            entity.click(&mouse);
            entity.stay_inside(&self.size);
        }
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pub struct Entity {
    kinetics: s5::KineticState,
    color: s5::Color,
    shape: Box<dyn Shape>,
    animator: Option<Box<dyn Animator>>,
    interactor: Option<Box<dyn Interactor>>,
}

impl Entity {
    pub fn new(
        kinetics: s5::KineticState,
        color: s5::Color,
        shape: Box<dyn Shape>,
        animator: Option<Box<dyn Animator>>,
        interactor: Option<Box<dyn Interactor>>,
    ) -> Self {
        Self {
            kinetics,
            color,
            shape,
            animator,
            interactor,
        }
    }

    pub fn kinetics(&self) -> &s5::KineticState {
        &self.kinetics
    }

    pub fn kinetics_mut(&mut self) -> &mut s5::KineticState {
        &mut self.kinetics
    }

    pub fn color(&self) -> s5::Color {
        self.color
    }

    pub fn color_mut(&mut self) -> &mut s5::Color {
        &mut self.color
    }

    pub fn bounding_radius(&self) -> f64 {
        self.shape.bounding_radius()
    }

    fn draw(
        &self,
        screen: &mut s5::Screen,
    ) {
        self.shape.draw(self, screen);
    }

    fn animate(
        &mut self,
        window_size: &s5::Rect,
        dt: f64,
    ) {
        self.kinetics.integrate(dt);
        if let Some(animator) = self.animator.take() {
            animator.animate(self, window_size, dt);
            self.animator = Some(animator);
        }
    }

    fn stay_inside(
        &mut self,
        window_size: &s5::Rect,
    ) {
        let kinetics = &mut self.kinetics;
        kinetics.x = kinetics.x.rem_euclid(window_size.w);
        kinetics.y = kinetics.y.rem_euclid(window_size.h);
    }

    fn click(
        &mut self,
        mouse: &s5::Point,
    ) {
        let radius = self.shape.bounding_radius();
        let dx = mouse.x - self.kinetics.x;
        let dy = mouse.y - self.kinetics.y;
        if dx * dx + dy * dy <= radius * radius {
            if let Some(interactor) = self.interactor.take() {
                interactor.click(self, mouse);
                self.interactor = Some(interactor);
            }
        }
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pub trait Shape {
    fn bounding_radius(&self) -> f64;
    fn draw(
        &self,
        entity: &Entity,
        screen: &mut s5::Screen,
    );
}

pub trait Animator {
    fn animate(
        &self,
        entity: &mut Entity,
        window_size: &s5::Rect,
        dt: f64,
    );
}

pub trait Interactor {
    fn click(
        &self,
        entity: &mut Entity,
        mouse: &s5::Point,
    );
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
